
#include "lib.h"

// TODO make it in a cleaner way
int indent = 0;
char tabulation[INDENT_BUF] = "";

#define INDENT_MAKER
struct symtable * symtable_new()
{
    struct symtable * t = malloc(sizeof(struct symtable));
    if(t == NULL){
        fprintf(stderr,"Bug memory allocation failed\n");
        exit(1);
    }
    t->capacity = 1024;
    t->symbols = malloc(t->capacity*sizeof(struct symbol));
    if(t->symbols  == NULL){
        fprintf(stderr,"Bug memory allocation failed\n");
        exit(1);
    }
    t->temporary = 0;
    t->size = 0;
    return t;
}

static void symtable_grow(struct symtable * t) 
{
    t->capacity += 1024;
    t->symbols = realloc(t->symbols,t->capacity*sizeof(struct symbol));
    if(t->symbols == NULL) 
    {
      fprintf(stderr,"Error attempting to grow symbol table (actual size is %d)\n",t->size);
        exit(1);
    }
}

struct symbol* symtable_const_int(struct symtable * t, long int v) 
{
    unsigned int i;
    for ( i=0 ; i<t->size; i++ ){
        if(t->symbols[i].kind == CONSTANT_INT){
            if(t->symbols[i].u.value_int == v){
                break;
            }
        }
    }
    if(i==t->size)
    {
        if(t->size==t->capacity)
          symtable_grow(t);
        struct symbol *s = &(t->symbols[t->size]);
        s->kind = CONSTANT_INT;
        s->u.value_int = v;
        ++ (t->size);
        return s;
    }
    else 
    {
        return &(t->symbols[i]);
    }
}

struct symbol* symtable_const_double(struct symtable * t, float v) 
{
    unsigned int i;
    for ( i=0 ; i<t->size; i++ ){
        if(t->symbols[i].kind == CONSTANT_FLOAT){
            if(t->symbols[i].u.value_float == v){
                break;
            }
        }
    }
    if(i==t->size)
    {
        if(t->size==t->capacity)
          symtable_grow(t);
        struct symbol *s = &(t->symbols[t->size]);
        s->kind = CONSTANT_FLOAT;
        s->u.value_float = v;
        ++ (t->size);
        return s;
    }
    else 
    {
        return &(t->symbols[i]);
    }
}

struct symbol* symtable_get(struct symtable * t, const char * id, struct symbol* func_id) 
{
    unsigned int i;
    for ( i=0 ; i<t->size && strcmp(t->symbols[i].u.name,id) != 0; i++ );
    if(i<t->size){
        if(t->symbols[i].kind == NAME_LOC || t->symbols[i].kind == ARRAY){
            if(strcmp(func_id->u.name,t->symbols[i].scope_function) == 0){
                return &(t->symbols[i]);
            }
            return NULL;
        }
        return &(t->symbols[i]);
    }
      
    return NULL;
}

struct symbol* symtable_put_name(struct symtable * t, const char * id) 
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = NAME;
    strcpy(s->u.name,id);
    ++ (t->size);
    return s;
}

struct symbol* symtable_put_chaine(struct symtable * t, const char * chaine) 
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = CHAIN;
    strcpy(s->u.chaine,chaine);
    ++ (t->size);
    return s;
}

struct symbol* symtable_put_loc(struct symtable * t, const char * var_name, struct symbol* func_id) 
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = NAME_LOC;
    strcpy(s->u.name,var_name);
    strcpy(s->scope_function,func_id->u.name);
    ++ (t->size);
    return s;
}

struct symbol* symtable_put_array(struct symtable * t, const char* var_name, unsigned int dim, struct symbol** arrayDim, struct symbol* dim1size, struct symbol* func_id){
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);

    s->kind = ARRAY;
    strcpy(s->u.arr.name,var_name);
    s->u.arr.dim = dim;
    s->u.arr.dim1size = dim1size;

    // Allocate a new array inside the structure because the argument arrayDim will be freed
    s->u.arr.size = (struct symbol**) malloc(dim*sizeof(struct symbol*));
    if(s->u.arr.size == NULL){
        fprintf(stderr,"Bug memory allocation failed\n");
        exit(1);
    }

    // Copy argument arrayDim in the array inside the structure
    for(unsigned int i=0;i<dim;i++){
        s->u.arr.size[i] = arrayDim[i];
    }

    strcpy(s->scope_function,func_id->u.name);

    ++ (t->size);
    return s;
}

void symtable_dump(struct symtable * t, FILE* fout)
{
    unsigned int i;
    for ( i=0 ; i<t->size; i++ )
    {
      if(t->symbols[i].kind==CONSTANT_INT)
        fprintf(fout,"#       %p = %ld\n",&(t->symbols[i]),t->symbols[i].u.value_int);
      else if(t->symbols[i].kind==CONSTANT_FLOAT)
        fprintf(fout,"#       %p = %0.2f\n",&(t->symbols[i]),t->symbols[i].u.value_float);
      else if(t->symbols[i].kind==NAME)
        fprintf(fout,"#       %p = %s\n",&(t->symbols[i]),t->symbols[i].u.name);
      else if(t->symbols[i].kind==NAME_LOC)
        fprintf(fout,"#       %p = %s (in the scope of the '%s' function)\n",&(t->symbols[i]),t->symbols[i].u.name,t->symbols[i].scope_function);
      else if(t->symbols[i].kind==CHAIN)
        fprintf(fout,"#       %p = %s\n",&(t->symbols[i]),t->symbols[i].u.chaine);
      else if(t->symbols[i].kind==ARRAY){
        fprintf(fout,"#       %p = (%s,",&(t->symbols[i]),t->symbols[i].u.arr.name);
        if(t->symbols[i].u.arr.size[0]->kind == NAME_LOC || t->symbols[i].u.arr.size[0]->kind == NAME)
          fprintf(fout,"%s",t->symbols[i].u.arr.size[0]->u.name);
        else if(t->symbols[i].u.arr.size[0]->kind == CONSTANT_INT)
          fprintf(fout,"%ld",t->symbols[i].u.arr.size[0]->u.value_int);
        else if(t->symbols[i].u.arr.size[0]->kind == CONSTANT_FLOAT)
          fprintf(fout,"%0.2f",t->symbols[i].u.arr.size[0]->u.value_float);
        else{
          printf("BUG array size kind:%d\n",t->symbols[i].u.arr.size[0]->kind); 
          exit(1);
        }
        fprintf(fout,") (in the scope of the '%s' function)\n",t->symbols[i].scope_function);
      }
    }
    fprintf(fout,"#       --------\n");;
}

void symtable_free(struct symtable * t)
{
    for (unsigned int i=0 ; i<t->size; i++ ){
        if(t->symbols[i].kind == ARRAY){
            if(t->symbols[i].u.arr.size != NULL)
                free(t->symbols[i].u.arr.size);
        }
    }
    free(t->symbols);
    free(t);
}

struct code * code_new()
{
    struct code * r = malloc(sizeof(struct code));
    r->capacity = 1024;
    r->quads = malloc(r->capacity*sizeof(struct quad));
    r->nextquad = 0;
    return r;
}

static void code_grow(struct code * c)
{
    c->capacity += 1024;
    c->quads = realloc(c->quads,c->capacity*sizeof(struct quad));
    if(c->quads == NULL) 
    {
      fprintf(stderr,"Error attempting to grow quad list (actual size is %d)\n",c->nextquad);
        exit(1);
    }
}

void gencode(struct code * c,
              enum quad_kind k,
              struct symbol* s1,
              struct symbol* s2,
              struct symbol* s3)
{
    if ( c->nextquad == c->capacity )
        code_grow(c);
    struct quad * q = &(c->quads[c->nextquad]);
    q->kind = k;
    q->sym1 = s1;
    q->sym2 = s2;
    q->sym3 = s3;
    ++ (c->nextquad);
}

struct symbol *newtemp(struct symtable * t)
{
  struct symbol* s;
  name_t name;
  sprintf(name,"t%d",t->temporary);
  s = symtable_put_name(t,name);
  ++ (t->temporary);
  return s;
}

static void symbol_dump(struct symbol* s, FILE* fout)
{
    switch ( s->kind )
    {
        case NAME:
            fprintf(fout,"%s",s->u.name);
            break;
        case CONSTANT_INT:
            fprintf(fout,"%ld",s->u.value_int);
            break;
        case CONSTANT_FLOAT:
            fprintf(fout,"%0.2f",s->u.value_float);
            break;
        case NAME_LOC:
            fprintf(fout,"%s",s->u.name);
            break;
        case CHAIN:
            fprintf(fout,"%s",s->u.chaine);
            break;
        case ARRAY:
            fprintf(fout,"%s",s->u.arr.name);
            break;
        default:
            fprintf(stderr,"BUG SYMBOL DUMP NOT RECOGNIZED\n");
            break;
    }
}

static void quad_dump(struct quad * q, FILE* fout)
{
    switch ( q->kind )
    {
        case BOP_PLUS:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," + ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_MINUS:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," - ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_MULT:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," * ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_DIV:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," / ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_MOD:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," %% ");
            symbol_dump(q->sym3,fout);
            break;
        case UOP_MINUS:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            fprintf(fout,"- ");
            symbol_dump(q->sym2,fout);
            break;
        case DEF_MAIN:
            symbol_dump(q->sym1,fout);
            fprintf(fout,":");
            break;
        case DEF_FUNCTION:
            symbol_dump(q->sym1,fout);
            fprintf(fout,":");
            break;
        case CALL_PRINT:
            fprintf(fout,"print ");
            symbol_dump(q->sym1,fout);
            break;
        case CALL_PRINTSTR:
            fprintf(fout,"printstr ");
            symbol_dump(q->sym1,fout);
            break;
        case COPY:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            break;
        case ALLOCATE_ARRAY:
            fprintf(fout,"Allocate '");
            symbol_dump(q->sym1,fout);
            fprintf(fout,"' to an one dimensional array of length ");
            symbol_dump(q->sym1->u.arr.dim1size,fout);
            break;
        case AFF_ARRAY:
            symbol_dump(q->sym1,fout);
            fprintf(fout,"[");
            symbol_dump(q->sym2,fout);
            fprintf(fout,"] := ");

            symbol_dump(q->sym3,fout);
            break;
        case COPY_ARRAY:
            symbol_dump(q->sym1,fout);
            
            fprintf(fout," := ");
            
            symbol_dump(q->sym2,fout);
            fprintf(fout,"[");
            symbol_dump(q->sym3,fout);
            fprintf(fout,"]");
            break;
        default:
            fprintf(stderr,"BUG quad_dump(...) : q->kind inconnu\n");
            break;
    }
}

void code_dump(struct code * c, FILE* fout)
{
    unsigned int i;
    for ( i=0 ; i<c->nextquad ; i++ )
    {   
        fprintf(fout,"#%4u \t| ",i);
        quad_dump(&(c->quads[i]),fout);
        fprintf(fout,"\n");
    }
    fprintf(fout,"\n\n");
}

int indice_temp(struct symtable* t, struct symbol* s){
    unsigned int i=0;
    for(i=i; i<t->size; i++){ 
        if(s->u.name == t->symbols[i].u.name){
            return i;
        }
    }
    fprintf(stderr,"Indice_temp: s n'est pas dans la table\n");
    exit(1);

    return i;
}

void code_mips_dump(struct symtable * t, struct code * c, FILE* fout){
    fprintf(fout,".data\n");
    fprintf(fout,"\tlf: .asciiz \"\\n\"\n");
    fprintf(fout,"\tspace: .asciiz \" \"\n");
    fprintf(fout,"\ttemp_float: .float 0.0\n");
    if(t==NULL){fprintf(stderr,"Code mips dump: symtable null\n"); exit(1);}
    
    unsigned int i =0;
    for(i=i; i<t->size; i++){ 
        if(t->symbols[i].kind == CONSTANT_INT){ //INTEGER CONSTANT 
            fprintf(fout,"\ttemp%d: .word %ld\n",i,t->symbols[i].u.value_int);
        }
        else if(t->symbols[i].kind == CONSTANT_FLOAT){ //FLOAT CONSTANT
            fprintf(fout,"\ttemp%d: .float %0.2f\n",i,t->symbols[i].u.value_float);
        }
        else if(t->symbols[i].kind == CHAIN){ //CHAIN
            fprintf(fout,"\ttemp%d: .asciiz %s\n",i,t->symbols[i].u.chaine);
        }
        else if(t->symbols[i].kind == NAME_LOC){
            fprintf(fout,"\t%s: .word 0\n",t->symbols[i].u.name);
        }
        else if(t->symbols[i].kind == NAME){
            fprintf(fout,"\t%s: .word 0\n",t->symbols[i].u.name);
        } 
        // this prevent the declaration fo some temp necessary but is also necessary to allocat the memory tab TODO resolve this issue
        else if (t->symbols[i].kind == ARRAY){
            fprintf(fout,"\t%s: .word 0\n",t->symbols[i].u.name);
        }
        else{ //NAME
            fprintf(fout,"\ttemp%d: .word 0\n",i);
        }
    }
    
    #ifdef LIBDEBUG
    debug_register_strings_declaration(fout);
    #endif

    print_macros(fout);
    
    // start of the .text segment
    fprintf(fout,"\n\n.text\n");

    for(i = 0; i<c->nextquad; i++){
        switch (c->quads[i].kind)
        {
            case BOP_PLUS:
                print_bop_plus(c,fout,i);
                break;
            case BOP_MINUS:
                print_bop_minus(c,fout,i);
                break;
            case BOP_MULT:
                print_bop_mult(c,fout,i);
                break;
            case BOP_DIV:
                print_bop_div(c,fout,i);
                break;
            case BOP_MOD:
                print_bop_mod(c,fout,i);
                break;
            case UOP_MINUS:
                print_uop_minus(c,fout,i);
                break;
            case COPY_ARRAY:
                print_copy_array(t,c,fout,i);
                break;
            case ALLOCATE_ARRAY:
                print_allocate_array(t,c,fout,i);
                break;
            case AFF_ARRAY:
                print_aff_array(t,c,fout,i);
                break;
            case DEF_MAIN:
                print_main_def(fout);
                add_indent(1,c,t,fout);
                break;
            case CALL_PRINT:
                print_print(t,c,fout,i);
                break;
            case CALL_PRINTSTR:
                print_printstr(t,c,fout,i);
                break;
            case COPY:
                print_copy(t,c,fout,i);
                break;
            default:
                fprintf(stderr,"#BUG %d\n",c->quads[i].kind);
                break;
        }
    }

    fprintf(fout,"\t# exit\n");
    fprintf(fout,"\tli $v0,10\n");
    fprintf(fout,"\tsyscall\n");
}

void code_free(struct code * c)
{
    free(c->quads);
    free(c);
}

void free_exit(int code, struct code* c, struct symtable* t, FILE* fname){
    code_free(c);
    symtable_free(t);
    if(fname!=NULL){fclose(fname);}
    exit(code);
}


/************************* MIPS CODE GEN FUNCTIONS *************************/

// -------- MAIN RELATED --------

void print_main_def(FILE* fout){
    fprintf(fout,".globl start\n\n");
    fprintf(fout,"start:\n");
}

// -------- ARITHMETIC RELATED --------

void print_bop_plus(struct code * c, FILE* fout,int i)
{
    //Check if we have variable names or just integers for sym2 and sym3
    print_nameOrInteger(c,fout,i,0b110);
    //Do the operation and save result in wanted variable
    fprintf(fout,"\tadd $t0, $t0, $t1\n");
    fprintf(fout,"\tsw $t0, %s\n",c->quads[i].sym1->u.name);
}

void print_bop_minus(struct code * c, FILE* fout,int i)
{
    //Check if we have variable names or just integers
    print_nameOrInteger(c,fout,i,0b110);
    //Do the operation and save result in wanted variable
    fprintf(fout,"\tsub $t0, $t0, $t1\n");
    fprintf(fout,"\tsw $t0, %s\n",c->quads[i].sym1->u.name);
}

void print_bop_mult(struct code * c, FILE* fout,int i)
{
    //Check if we have variable names or just integers for sym2 and sym3
    print_nameOrInteger(c,fout,i,0b110);
    //Do the operation and save result in wanted variable
    fprintf(fout,"%smul $t0, $t0, $t1\n",tabulation);
    fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
}

void print_bop_div(struct code * c, FILE* fout,int i)
{
    //Check if we have variable names or just integers for sym2 and sym3
    print_nameOrInteger(c,fout,i,0b110);
    //Do the operation and save result in wanted variable
    fprintf(fout,"%sdiv $t0, $t0, $t1\n",tabulation);
    fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
}

void print_bop_mod(struct code * c, FILE* fout,int i)
{
    //Check if we have variable names or just integers for sym2 and sym3
    print_nameOrInteger(c,fout,i,0b110);
    //Do the operation and save result in wanted variable
    fprintf(fout,"%sdiv $t0, $t0, $t1\n",tabulation);
    fprintf(fout,"%smfhi $t0\n",tabulation);
    fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
}

void print_uop_minus(struct code * c, FILE* fout,int i)
{
    fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    //Check if we have variable names or just integers
    if(c->quads[i].sym2->kind == NAME || c->quads[i].sym2->kind == NAME_LOC)
        fprintf(fout,"%slw $t1, %s\n",tabulation,c->quads[i].sym2->u.name);
    else if(c->quads[i].sym2->kind == CONSTANT_INT)
        fprintf(fout,"%sli $t1, %ld\n",tabulation,c->quads[i].sym2->u.value_int);

    //Do the operation and save result in wanted variable
    fprintf(fout,"%ssub $t0, $t0, $t1\n",tabulation);
    fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
}

// -------- PRINT RELATED --------

void print_print(struct symtable * t,struct code * c, FILE* fout,int i)
{
    switch(c->quads[i].sym1->kind){
        case ARRAY:
            // TODO + WIP solution temporaire, quand les fonctions seront implémentés à remplacer par une fonction pour éviter le code spaghetti
            //  putting the array size in $t2
            load_symbol_macro("$t2",c->quads[i].sym1->u.arr.dim1size,t,fout);
            // to avoid doing a mult at each iterations we multiply the size by 4 (each elements need a 4 bytes incrementation)
            fprintf(fout,"%smul $t0, $t2, 4\n",tabulation);
            //  putting array address in $t1
            load_symbol_macro("$t1",c->quads[i].sym1,t,fout);
            fprintf(fout,"%sli $t2, 0\n",tabulation);


            // we avoid having 2 identical labels by putting the quad index at the end of it
            // /!\ start of the loop /!\ --
            fprintf(fout,"%sprint_array_loop%d:\n",tabulation,i);
            add_indent(1,c,t,fout);
            // bge by security to avoid miss matching (which shouldn't be happening)
            fprintf(fout,"%sbge $t2, $t0, end_array_loop%d\n",tabulation,i);
            
            // put the address of the current element to be printed in $t3
            fprintf(fout,"%sadd $t3, $t2, $t1\n",tabulation);
            // load the current float element in the
            fprintf(fout,"%slwc1 $f0, ($t3)\n",tabulation);
            // function that print the asked register
            print_reg("f0",fout,1);
            // macro that print a space
            fprintf(fout,"%sSPACE\n",tabulation);
            // increment to next address
            fprintf(fout,"%saddi $t2, $t2, 4\n",tabulation);

            // jump to the start of the loop
            fprintf(fout,"%sj print_array_loop%d\n",tabulation,i);
            // /!\ end of the loop /!\ --
            remove_indent();
            fprintf(fout,"%send_array_loop%d:\n",tabulation,i);
            fprintf(fout,"%sLF\n",tabulation);
            break;
        default:
            fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
            fprintf(fout,"%smove $a0, $t0\n",tabulation);
            // ne marche que pour print ls int pas pour les float
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            fprintf(fout,"%ssyscall\n",tabulation);
            //Return to line
            fprintf(fout,"%sLF\n",tabulation);
            break;
    }
    
}

void print_printstr(struct symtable * t, struct code * c, FILE* fout,int i)
{
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, temp%d\n",tabulation,indice_temp(t,c->quads[i].sym1));
    fprintf(fout,"%ssyscall\n",tabulation);
}


void print_reg(char * reg,FILE * fout, int type){
    switch(type) {
        case 0:
            fprintf(fout,"%smove $a0, $%s\n",tabulation,reg);
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
        case 1:
            fprintf(fout,"%smov.s $f12, $%s\n",tabulation,reg);
            fprintf(fout,"%sli $v0, 2\n",tabulation);
            break;
        default:
            // we suppose value is int
            fprintf(fout,"%smove $a0, $%s\n",tabulation,reg);
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
    }
    fprintf(fout,"%ssyscall\n",tabulation);
}

// -------- FUNCTIONS RELATED --------


// -------- ARRAY RELATED --------

void print_copy_array(struct symtable * t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# copy_array section\n",tabulation);
    #endif
    if(c->quads[i].sym2->kind != ARRAY){
        fprintf(stdout,"Tried to do an  Var = tab[index] operation but right part isn't an array should address this issue in the corec.y file\n");
        exit(1);
    }
    // adress of the array element we want to copy
    load_tab_addr_macro("$t0", 1, 2, c->quads[i].sym2,c->quads[i].sym3,t,fout);
    // adress of the place we want to copy the array element in
    load_addr_macro("$t1",c->quads[i].sym1,t,fout);
    // load in $t2 what's inside the array element
    fprintf(fout,"%slw $t2, ($t0)\n",tabulation);
    // store it in the destination
    fprintf(fout,"%ssw $t2, ($t1)\n",tabulation);

    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of copy_array section\n",tabulation);
    #endif
}

void print_allocate_array(struct symtable * t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# allocate_array section\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind != ARRAY){
        fprintf(stdout,"Tried to don an tab[index] = Var operation but left part isn't an array should address this issue in the corec.y file\n");
        exit(1);
    }
    // number for the sbrk primitive call
    fprintf(fout,"%sli $v0, 9\n",tabulation);
    // based on the type of the var indicating the size of the array we do different operations to end up with the nb of bytes we should allocate is $a0
    load_symbol_alloc_macro("$a0",0,c->quads[i].sym1->u.arr.dim1size,t,fout);
    // call sbrk
    fprintf(fout,"%ssyscall\n",tabulation);
    // save the allocated array ptr returned in $v0 by syscall in the assigned .data var
    fprintf(fout,"%sla $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    fprintf(fout,"%ssw $v0, ($t0)\n",tabulation);
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of allocate_array section\n",tabulation);
    #endif
}

void print_aff_array(struct symtable * t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# aff_array section\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind != ARRAY){
        fprintf(stdout,"Tried to do an tab[index] = Var operation but left part isn't an array should address this issue in the corec.y file\n");
        exit(1);
    }
    // V alternative, effectue directement le*4 en calcul vu qu'on devrait passer par un load immediate dans tout les cas 
    // fprintf(fout,"%sla $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    load_tab_addr_macro("$t0", 1, 2, c->quads[i].sym1,c->quads[i].sym2,t,fout);
    load_symbol_macro("$t1",c->quads[i].sym3,t,fout);
    fprintf(fout,"%ssw $t1, ($t0)\n",tabulation);
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of aff_array section\n",tabulation);
    #endif
}   




// -------- MISCELLANEOUS --------

void print_copy(struct symtable * t, struct code * c, FILE* fout,int i)
{
    if(c->quads[i].sym2->kind == NAME || c->quads[i].sym2->kind == NAME_LOC)
        fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym2->u.name);
    else
        fprintf(fout,"%slw $t0, temp%d\n",tabulation,indice_temp(t,c->quads[i].sym2));
    fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
}

/***    do the right thing based on whether we have an named var or a constant TODO finde a better name !
 *      if on sym 1 sym = 0b1 do on sym1 (not implmented should never be of use)
 *      if on sym 2 sym = 0b10 etc
 *      if on sym 2 and 3 but not 1 simultaneously sym = 0b110
 *      to use only on BOP operations PLUS MINUS MUL DIV MOD
 *      TODO sym may be useless and may be removed later
 */ 

void print_nameOrInteger(struct code * c, FILE* fout,int i,int sym)
{
    if(sym & 0b10) {
        if(c->quads[i].sym2->kind == NAME || c->quads[i].sym2->kind == NAME_LOC)
            fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym2->u.name);
        else if(c->quads[i].sym2->kind == CONSTANT_INT)
            fprintf(fout,"%sli $t0, %ld\n",tabulation,c->quads[i].sym2->u.value_int);
    }
    if( sym & 0b100) {
        if(c->quads[i].sym3->kind == NAME || c->quads[i].sym3->kind == NAME_LOC)
            fprintf(fout,"%slw $t1, %s\n",tabulation,c->quads[i].sym3->u.name);
        else if(c->quads[i].sym3->kind == CONSTANT_INT)
            fprintf(fout,"%sli $t1, %ld\n",tabulation,c->quads[i].sym3->u.value_int);
    }
}


// -------- MACROS --------
void load_symbol_alloc_macro(char * dest, int tmp_used, struct symbol * sym, struct symtable * t, FILE* fout){
    if(sym->kind == CONSTANT_INT) // var = 3
        // in this case the value is a constant int we can do a load immediate in $a0 with value*4
        fprintf(fout,"%sli %s, %ld\n",tabulation,dest,(sym->u.value_int)*4);
    // in other cases we need to load the value from the label of the corresponding symbol
    else if( sym->kind == NAME || sym->kind == NAME_LOC ) { // var = 4*3
        fprintf(fout,"%slw $t%d, %s\n",tabulation,tmp_used,(sym->u.name));
    } else { // array[4] = array[5]+3*4
        fprintf(fout,"%slw $t%d, temp%d\n",tabulation,tmp_used,indice_temp(t,sym));
    }
    // if we don't have a CONST INT we loaded the nb of int/floats we want to stock in the array, in this case
    if(sym->kind  != CONSTANT_INT)
        fprintf(fout,"%ssll %s, $t%d, 2\n",tabulation,dest,tmp_used);
}

void load_symbol_macro(char * dest, struct symbol * sym, struct symtable * t, FILE* fout){
    if(sym->kind == CONSTANT_INT) // var = 3
        // in this case the value is a constant int we can do a load immediate in $a0 with value*4
        fprintf(fout,"%sli %s, %ld\n",tabulation,dest,(sym->u.value_int));
    else if(sym->kind == NAME_LOC){
        fprintf(fout,"%slw %s, %s\n",tabulation,dest,(sym->u.name));
        fprintf(fout,"%smtc1 %s, $f0\n",tabulation,dest);
        fprintf(fout,"%scvt.s.w $f2, $f0\n",tabulation);
        fprintf(fout,"%sswc1 $f2, temp_float\n",tabulation);
        fprintf(fout,"%slw %s, temp_float\n",tabulation,dest);
    }
    // in other cases we need to load the value from the label of the corresponding symbol
    else if( sym->kind == NAME || sym->kind == ARRAY) { // var = 4*3
        fprintf(fout,"%slw %s, %s\n",tabulation,dest,(sym->u.name));
    } else { // array[4] = array[5]+3*4
        fprintf(fout,"%slw %s, temp%d\n",tabulation,dest,indice_temp(t,sym));
    }
}


void load_tab_addr_macro(char * dest,int tmp_used1, int tmp_used2, struct symbol * sym_addr, struct symbol * sym_index, struct symtable * t, FILE* fout){
    fprintf(fout,"%slw $t%d, %s\n",tabulation,tmp_used2,sym_addr->u.name);
    if(sym_index->kind == CONSTANT_INT)
        // put into dest the addr of array[index] element
        fprintf(fout,"%saddi %s, $t%d, %ld\n",tabulation,dest,tmp_used2,(sym_index->u.value_int)*4);
    // in other cases we need to load the value from the label of the corresponding symbol
    else if( sym_index->kind == NAME || sym_index->kind == NAME_LOC || sym_index->kind == ARRAY) {
        fprintf(fout,"%slw $t%d, %s\n",tabulation,tmp_used1,(sym_index->u.name));
    } else {
        fprintf(fout,"%slw $t%d, temp%d\n",tabulation,tmp_used1,indice_temp(t,sym_index));
    }
    if (sym_index->kind != CONSTANT_INT){
        // if not CONSTANT_INT wasn't yet multiplied by 4
        fprintf(fout,"%ssll $t%d, $t%d, 2\n",tabulation,tmp_used1,tmp_used1);
        fprintf(fout,"%sadd %s, $t%d, $t%d\n",tabulation,dest,tmp_used2,tmp_used1);
    }
}

void load_addr_macro(char * dest, struct symbol * sym, struct symtable * t, FILE* fout)
{
    if( sym->kind == NAME || sym->kind == NAME_LOC || sym->kind == ARRAY){
        fprintf(fout,"%sla %s, %s\n",tabulation,dest,sym->u.name);
    } else 
        fprintf(fout,"%sla %s, temp%d\n",tabulation,dest,indice_temp(t,sym));

}


void print_macros(FILE* fout){
    fprintf(fout,"\n# MACROS definition\n\n");
    print_LF_macro(fout);
    print_SPACE_macro(fout);
    fprintf(fout,"# End of MACROS definition\n");
}

void print_LF_macro(FILE* fout){
    fprintf(fout,".macro LF\n");
    fprintf(fout,"\tli $v0, 4\n");
    fprintf(fout,"\tla $a0, lf\n");
    fprintf(fout,"\tsyscall\n");
    fprintf(fout,".end_macro\n");
}

void print_SPACE_macro(FILE* fout){
    fprintf(fout,".macro SPACE\n");
    fprintf(fout,"\tli $v0, 4\n");
    fprintf(fout,"\tla $a0, space\n");
    fprintf(fout,"\tsyscall\n");
    fprintf(fout,".end_macro\n");
}
// ETC FEEL FREE TO ADD MORE SECTIONS

// -------- DEBUG RELATED --------

#ifdef LIBDEBUG
void debug_print_reg(char * reg,FILE * fout, int type){
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, print%s\n",tabulation,reg);
    fprintf(fout,"%ssyscall\n",tabulation);
    
    fprintf(fout,"%smove $a0, $%s\n",tabulation,reg);
    switch(type) {
        case 0:
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
        case 1:
            fprintf(fout,"%sli $v0, 4\n",tabulation);
            break;
        default:
            // we suppose value is int
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
    }
    fprintf(fout,"%ssyscall\n",tabulation);
    //Return to line
    fprintf(fout,"%sLF\n",tabulation);

}

void debug_print_start(FILE * fout){
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, startfct\n",tabulation);
    fprintf(fout,"%ssyscall\n",tabulation);
    fprintf(fout,"%sLF\n",tabulation);
}

void debug_print_end(FILE * fout){
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, endfct\n",tabulation);
    fprintf(fout,"%ssyscall\n",tabulation);
    fprintf(fout,"%sLF\n",tabulation);
}

void debug_register_strings_declaration(FILE * fout){
    for(int i = 0; i < 10; i++)
        fprintf(fout,"\tprintt%d: .asciiz \"$t%d : \"\n",i,i);
    for(int i = 0; i < 4; i++)
        fprintf(fout,"\tprinta%d: .asciiz \"$a%d : \"\n",i,i);
    for(int i = 0; i < 3; i++)
        fprintf(fout,"\tprintv%d: .asciiz \"$v%d : \"\n",i,i);
    fprintf(fout,"\tstartfct: .asciiz \"START \"\n" );
    fprintf(fout,"\tendfct: .asciiz \"END \"\n" );
    
}
#endif

/************************* NAME TO DEFINE FUNCTIONS *************************/



void add_indent(int code, struct code* c, struct symtable* t, FILE* fname){
    if(indent >= 1022)
    {
        printf("more than 1023 indentations ! write better code.\n");
        free_exit(code,c,t,fname);
    }
    tabulation[indent] = '\t';
    indent++;
    tabulation[indent] = '\0';
}

int remove_indent(){
    if(indent == 0)
    {
        printf("shouldn't be removing a tab since there are none, resuming \n");
        return -1;
    }
    indent--;
    tabulation[indent] = '\0';
    return 0;
}