#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

extern unsigned int lineNumber;
extern struct symbol* symb_scope_function;

#define INDENT_BUF 1024
// make each section of MIPS code more clear comment to disable
#define LIBDEBUG

/* TABLE DES SYMBOLES */

typedef char name_t[64];
typedef char name_64_t[1024];

struct array {
    name_t name;
    struct symbol** size;
    unsigned int dim;
    struct symbol* dim1size;
};

enum kind{ 
    NAME,             // -> use the name attribut in the union u
    CONSTANT_INT,     // -> use the value_int attribut in the union u
    CONSTANT_FLOAT,   // -> use the value_float attribut in the union u
    CHAIN,            // -> use the chaine attribut in the union u
    NAME_LOC,         // -> use the name attribut, the symbol is just defined
    ARRAY             // -> use the name attribut in the union u
};

// Struct symbol -> store a symbol
struct symbol {
    enum kind kind; // attribut kind : the type of symbol;
    union {
        name_t name;
        name_64_t chaine;
        long int value_int;
        float value_float;
        struct array arr;
    } u;
    // * (! only sets with variable kind : NAME_LOC, ARRAY)
    name_t scope_function; // -> Name of the function where locals variables (NAME_LOC, ARRAY) are defined. 
};

// Store the symbol table
struct symtable {
    unsigned int capacity;
    unsigned int temporary; 
    unsigned int size;
    struct symbol* symbols;
};

// Initiate the symtable
struct symtable * symtable_new();

// Add a symbol in the table of a integer (CONSTANT_INT)
struct symbol* symtable_const_int(struct symtable * t, long int v);
// Add a symbol in the table of a float (CONSTANT_FLOAT)
struct symbol* symtable_const_double(struct symtable * t, float v);
// Add a symbol in the table of a function (NAME)
struct symbol* symtable_put_name(struct symtable * t, const char * s);
// Add a symbol in the table of a chaine of caracters (CHAIN)
struct symbol* symtable_put_chaine(struct symtable * t, const char * chaine);
// Add a symbol in the table of a local variable (defined but isn't initialized) (NAME_LOC)
struct symbol* symtable_put_loc(struct symtable * t, const char * var_name, struct symbol* func_id);
// Add a symbol in the table of a local array (defined but isn't initialized) (ARRAY)
struct symbol* symtable_put_array(struct symtable * t, const char* var_name, unsigned int dim, struct symbol** arrayDim, struct symbol* dim1size, struct symbol* func_id);

// Get a pointer of a symbol with attribut name in the symtable (if 's' isn't in the table, return NULL)
// (func_id is used for the scope of function on the local variable)
struct symbol* symtable_get(struct symtable * t, const char * s, struct symbol* func_id);

// Display the symtable in the file output 'fout'
void symtable_dump(struct symtable * t, FILE* fout);
// Free the memory of the symtable
void symtable_free(struct symtable * t);


/* QUADRUPLETS ET CODE */
//#define _NB_BYTES 4

/*
* Struct quad -> store the an instruction
* BOP_* -> Binary operation
* UOP_* -> Unaire operation
* COPY -> AFFECTATION of variable (except array)
* AFF_ARRAY -> array=sym1 element=sym2 tmp=sym3 -> array[element]= tmp
* COPY_ARRAY -> tmp=sym1 array=sym2 element=sym3  -> tmp= array[element]
*/
struct quad {
    enum quad_kind { BOP_PLUS, BOP_MINUS, BOP_MULT, BOP_DIV, BOP_MOD, UOP_MINUS, COPY, COPY_ARRAY, ALLOCATE_ARRAY, AFF_ARRAY, DEF_MAIN, DEF_FUNCTION, CALL_PRINT, CALL_PRINTSTR} kind;
    struct symbol* sym1;
    struct symbol* sym2;
    struct symbol* sym3;
};

// Struct code -> store the intermediate code
struct code {
    unsigned int capacity;
    unsigned int nextquad;
    struct quad * quads;
};

// Initiate the struct code
struct code * code_new();

// Generate 3 addresses intermediate code
void gencode(struct code * c,
              enum quad_kind k,
              struct symbol* s1,
              struct symbol* s2,
              struct symbol* s3);

// Add to the symtable a temporary function which is used to transform multiples addresses instruction to a group of 3 addresses instructions (uses NAME kind)
struct symbol *newtemp(struct symtable * t);

// Display in the 'fout' output file the intermediate code (in commentary)
void code_dump(struct code * c, FILE* fout);
// Display in the 'fout' output file the mips code
void code_mips_dump(struct symtable * t, struct code * c, FILE* fout);

// Free memory of the code struct
void code_free(struct code * c);

void free_exit(int code, struct code* c, struct symtable* t, FILE* fname);


/************************* MIPS CODE GEN FUNCTIONS *************************/

// -------- MAIN RELATED --------

// print the starting flag so MIPS code know where he should start reading code from
void print_main_def(FILE* fout);

// -------- ARITHMETIC RELATED --------

// add the BOP plus arithmetic operation TODO explain in more details
void print_bop_plus(struct code * c, FILE* fout,int i);

// add the BOP minus arithmetic operation TODO explain in more details
void print_bop_minus(struct code * c, FILE* fout,int i);

// add the BOP mult arithmetic operation TODO explain in more details
void print_bop_mult(struct code * c, FILE* fout,int i);

// add the BOP div arithmetic operation TODO explain in more details
void print_bop_div(struct code * c, FILE* fout,int i);

// add the BOP modulo arithmetic operation TODO explain in more details
void print_bop_mod(struct code * c, FILE* fout,int i);

// add the UOP minus arithmetic operation TODO explain in more details
void print_uop_minus(struct code * c, FILE* fout,int i);

// -------- PRINT RELATED --------

// add the code to print the asked variable
void print_print(struct symtable * t, struct code * c, FILE* fout,int i);

// add the code to print the asked string
void print_printstr(struct symtable * t, struct code * c, FILE* fout,int i);

/** @brief print_reg make MIPS code to print the register's value, no LF and no SPACE
 *  only the register's value type = 0 => int | type = 1 => float everything else is considered as an int TODO update this
 *  TODO think about deleting type and just checking if the first reg character is an f if no other cases than int and floats
 */ 
void print_reg(char * reg,FILE * fout, int type);

// -------- FUNCTIONS RELATED --------

// -------- ARRAY RELATED --------

/** @brief print_copy_array make MIPS code that copy to a temp the value in an array at a given index X <= array[i]
 */
void print_copy_array(struct symtable * t, struct code * c, FILE* fout,int i);

/** @brief print_allocate_array make MIPS code to allocate the memory of a given array
 */
void print_allocate_array(struct symtable * t, struct code * c, FILE* fout,int i);

/** @brief print_aff_array make MIPS code to affect a value in an array at a given index array[i]= X
 */
void print_aff_array(struct symtable * t, struct code * c, FILE* fout,int i);

// -------- MISCELLANEOUS --------


// make a copy of a var TODO make a real comment
void print_copy(struct symtable * t, struct code * c, FILE* fout,int i);

// load whether a named var or a constant TODO find a better name !
void print_nameOrInteger(struct code * c, FILE* fout,int i,int sym);

// -------- MACROS --------

/** @brief load_symbol_alloc_macro load a symbol used for an sbrk allocation ( multiply it's value by 4)
 */
void load_symbol_alloc_macro(char * dest, int tmp_used, struct symbol * sym, struct symtable * t, FILE* fout);

/** @brief load_symbol_macro load a symbol value into the dest register
 */
void load_symbol_macro(char * dest, struct symbol * sym, struct symtable * t, FILE* fout);

/** @brief load_tab_addr_macro given a symbol adress and a symbol index put the addr of array[n] element in dest
 *  tmp_used1 and tmp_used2 won't be saved so ensure nothing that should be saved is in
 */
void load_tab_addr_macro(char * dest, int tmp_used1, int tmp_used2, struct symbol * sym_addr, struct symbol * sym_index, struct symtable * t, FILE* fout);

/** @brief load_addr_macro put the memory adress of sym into the dest register
 */
void load_addr_macro(char * dest, struct symbol * sym, struct symtable * t, FILE* fout);

/** @brief Macro printer only print the LF and SPACE macros for the moment
 * should be called between .data and .text segments
 */
void print_macros(FILE* fout);

/** @brief print_LF_macro add a macro to print a LF in MIPS code
 */
void print_LF_macro(FILE* fout);

/** @brief print_SPACE_macro add a macro to print a SPCE in MIPS code
 */
void print_SPACE_macro(FILE* fout);

// -------- DEBUG RELATED --------

#ifdef LIBDEBUG 
void debug_print_reg(char * reg,FILE * fout, int type);
void debug_print_start(FILE * fout);
void debug_print_end(FILE * fout);
void debug_register_strings_declaration(FILE * fout);
#endif
// ETC FEEL FREE TO ADD MORE SECTIONS
/************************* NAME TO DEFINE FUNCTIONS *************************/

// add a tabulation to the indentation string
void add_indent(int code, struct code* c, struct symtable* t, FILE* fname);

// remove a tabulation from the indentation string
int remove_indent();

